/**
 * =====================================================================================
 *        Class:  molecules
 *  Description:  contains information about molecules
 * =====================================================================================
 */
class molecules
{
  public: 
  size_t mol_numb;  
  virtual void plus_mol()=0;
  virtual void minus_mol()=0; 
  virtual ~molecules(){};
  molecules(size_t);
};
molecules::molecules(size_t i_mol_numb)
         : mol_numb(i_mol_numb)
  {}
////////////////////////////////////////////////////////////////////////////////////////
class mfm_molecules : public molecules
{
  public:
  void plus_mol() {++mol_numb;}
  void minus_mol(){--mol_numb;}
  mfm_molecules(size_t );
};
mfm_molecules::mfm_molecules(size_t init_mol_numb)
         : molecules(init_mol_numb)
  {}
/////////////////////////////////////////////////////////////////////////////////////////  
typedef list<molecule>::iterator   i_m;
class diff_molecules : public molecules
{
public:
  list<i_m>            mol_iters;    //iterators to positions of the molecule
  list<i_m>::iterator  iter_mi;      //iterator to list of iterators
  int                  mol_r;        //id of reacting molecule
  i_m                  h;           
  
  static gsl_rng*   _rand;       

  void plus_mol() {++mol_numb;}          
  void minus_mol(){--mol_numb;}       
  diff_molecules (size_t);
  i_m take_a_mol_iter();
  void put_a_mol_iter(i_m _it);
  static void put_rand(gsl_rng* r);
};
diff_molecules::diff_molecules(size_t init_mol_numb)
              : molecules(init_mol_numb)
  {}
gsl_rng* diff_molecules::_rand       = NULL;
i_m diff_molecules::take_a_mol_iter()
  {
    mol_r=gsl_rng_uniform_int(_rand,mol_numb);
    iter_mi=mol_iters.begin();
    for(int i(0);i<mol_r;++i) ++iter_mi;
    h=(*iter_mi);
//    mol_iters.erase(iter_mi);
    return(h);  
  }
void diff_molecules::put_a_mol_iter(i_m _it)
  {
    mol_iters.push_back(_it);   
  }
void diff_molecules::put_rand(gsl_rng* r)
  {
    _rand=r;
  }
/////////////////////////////////////////////////////////////////////////////////////////
class save_state
{
  public: 
  static double*      time;          ///<time from class simulation 
  virtual void write_state()=0; 
  virtual ~save_state(){};
};
/////////////////////////////////////////////////////////////////////////////////////////
double*  save_state::time        = NULL;
class mfm_molecules_write_conc : public mfm_molecules,save_state
{
  public:
  ofstream*           of;                    
  double              s_time;    // start time 
  double              p_time;    // previous time 
  double              e_time;    // end      time
  bool                with_step;
  double              step_time; // step     time
  mfm_molecules_write_conc(size_t i_mol_numb,double i_s_time,double i_e_time);
  mfm_molecules_write_conc(size_t i_mol_numb,double i_s_time,double i_e_time,double i_step_time);
  void plus_mol(); 
  void minus_mol(); 
  void complete_state();
  void write_initial_state();
  void complete();
  void write_state();
};
mfm_molecules_write_conc::mfm_molecules_write_conc(size_t i_mol_numb,double i_s_time,double i_e_time)
                        :mfm_molecules(i_mol_numb),s_time(i_s_time),p_time(s_time),e_time(i_e_time),with_step(false)
  {}
mfm_molecules_write_conc::mfm_molecules_write_conc(size_t i_mol_numb,double i_s_time,double i_e_time,double i_step_time)
                        :mfm_molecules(i_mol_numb),s_time(i_s_time),p_time(s_time),e_time(i_e_time),step_time(i_step_time),with_step(true)
  {}
void mfm_molecules_write_conc::complete_state()
  {
    if(with_step)
    {
      for( ; p_time <= e_time ; p_time += step_time)
      {
        (*of)<<p_time<<" "<<mol_numb<<"\n"; 
      }
    }
  }
void mfm_molecules_write_conc::write_initial_state()
  {
    if(!with_step)
    {    
        (*of)<<*time<<" "<<mol_numb<<"\n";
    }
  }
void mfm_molecules_write_conc::complete()
  {
     for( ; p_time < (*time) && p_time <=e_time; p_time += step_time)
     {
       (*of)<<p_time<<" "<<mol_numb<<"\n"; 
     }
  }
void mfm_molecules_write_conc::write_state()
  {
     if((*time) >= s_time && (*time) <= e_time)
     {    
       (*of)<<(*time)<<" "<<mol_numb<<"\n";
     }
  }
void mfm_molecules_write_conc::plus_mol() 
{
 if(with_step) 
 {
   complete();
   ++mol_numb;
 }
 else
 {
   ++mol_numb;
   write_state();
 }
}     
void mfm_molecules_write_conc::minus_mol() 
{
 if(with_step) 
 {
   complete();
   --mol_numb;
 }
 else
 {
   --mol_numb;
   write_state();
 }
}
/////////////////////////////////////////////////////////////////////////////////////////
class mfm_molecules_write_stopwatch : public mfm_molecules,save_state
{
  public:
  ofstream*           of;                    
  int                 e_n;         // end number
  bool                finish;

  mfm_molecules_write_stopwatch(size_t i_mol_numb, int i_e_n);
  void write_state();
  void plus_mol(); 
  void minus_mol(); 
};
mfm_molecules_write_stopwatch::mfm_molecules_write_stopwatch(size_t i_mol_numb, int i_e_n)
                             :mfm_molecules(i_mol_numb),e_n(i_e_n),finish(false)
  {}
void mfm_molecules_write_stopwatch::write_state()
{
    if(mol_numb == e_n)
    {
      (*of)<<(*time)<<endl;
      finish=true;
    }
}
void mfm_molecules_write_stopwatch::plus_mol() 
{
  ++mol_numb;
  write_state();
}     
void mfm_molecules_write_stopwatch::minus_mol() 
{
  --mol_numb;
  write_state();
}
