/**
  ========================================================================================\n
         Class:  parser     \n                                     
   Description:  Read input file and get informations about reactions and reagents.\n
         Usage:  All comments and white spaces are erased.\n
                 ;---comments \n
                 :---delimiters \n
                 c---reagents \n
                 The order is as follows: name of the molecule, number of molecules(required)
                 (n= ) or diffusion constant(optional) (d= ),radius of diffusive molecule (rad= )
                 and position (pos=[filename with positions or random])\n
                 r---reactions \n
                 Reactions should be written in the following manner (A+B>C+D).\n
                 Reversible reactions have to be written as two separate reactions.\n
                 Reaction constant(required) (k= ).\n
                 s---systems \n
                 At the first position a container should be given, for example sphere.\
                 Volume of the container (V= ).\
                 All informations are stored in vectors reactions and reagents as strings.\n
  =======================================================================================
*/
///\todo 
class parser
{
  private:
  ifstream&            input;     
  ofstream*            output;
  vector<string>       main_line;   ///< vector with tokens from crude_line
  vector<string>       inner_line;  ///< vector with tokens from main_line elements
  vector<string>       helper;      ///< vector with tokens from inner_line elements
  
  vector<string>       substr;      ///< helper for substrates        
  vector<string>       produc;      ///< helper for products          
  vector<string>       k;           ///< helper for k                 
                                         
  vector<vector<string> >      h;   ///< helper for reactions (elements of vector reactions)
  
  string               crude_line;  ///< crude line from input
  string               temp;        ///< helper

  const string         colon;       ///< ":"
  const string         plus;        ///< "+"
  const string         eq;          ///< "="
  const string         arrow;       ///< ">"
  const string         comma;       ///< ","
  const string         comment;     ///< ";"
  const string         num_mol;     ///< "n"
  const string         diff_const;  ///< "d"
  const string         react_const; ///< "k"
  const string         numb_step_del;///< "N" 
  const string         reagent;     ///< "c"
  const string         reaction;    ///< "r"
  const string         group;       ///< "g"
  const string         diff;        ///< "d"
  const string         null_reagent;///< "0"
  const string         radius;      ///< "rad"
  const string         sys;         ///< "s"
  const string         position;    ///< "pos"
  const string         volume;      ///< "V"
  const string         container;   ///< "sphere" 
  const string         division;    ///< "d"
  const string         result;      ///< "o"
  const string         numb_exp;    ///< number of experiments
  const string         time;        ///< "time"
  const string         from_time;   ///<"from_time";
  
  int                  line_number; ///< helper for logs(errors)
  bool                 exit_program;///< true --- mistake(s) in input file
  void wr_comm(string str);                           ///< write comments to logs
  void parse_line(const string&, vector<string>&, const string& delimiters = " "); ///< parse line
  bool check_if(const string& sth);
  void check_if_empty(vector<string>&);               ///< check if vector of srtings is empty (if no clear it)
  void check_if_empty(vector<vector<string> >&);      ///< check if vector of vectors is empty (if no clear it)
  bool check_if_reagent();                            ///< check reagent
  bool check_if_reaction();                           ///< check reaction
  void erase_white_spaces();                          ///< erase white spaces
  void erase_comments();                              ///< erase comments
  bool get(const string& wht,string& sth);
  bool get_mol_numb(string&);                         ///< get number of molecules 
  bool get_diff_const(string&);                       ///< get diffusion constant 
  bool get_mol_name(string&);                         ///< get name of molecule  
  bool get_react_const();                             ///< get k                
  bool get_reaction();                                ///< get substrate(s) and product(s), and write them to vectors
                                                      ///  substr and produc (false - no reaction or reagents)
  bool get_group();
  bool get_diffusive_reactions();
  bool get_concentrations();
  bool get_stopwatch();  
  bool get_radius(string&);
  bool get_position(string&);
  bool get_volume(string&);
  public:
  vector<vector<string> >           reagents;     /*!< info abouts reagents
  [0]---name;
  [1]---number;
  [2]---diff_const;
  [3]---radius;
  [4]---position*/
  vector<vector<string> >           diffusive_reactions;  /*!< info abouts diffusive particles and reactions*/
  vector<vector<string> >           groups_of_particles;  /*!< info abouts groups of particles*/
  vector<vector<vector<string> > >  reactions;    /*!< info abouts reactions
  [0]---substrate(s);
  [1]---product(s);
  [2]---reaction_const*/
  vector<vector<string> >           measurement; /*!< info about what is measured*/ 
  vector<string>                    system;      /*!< info about system
  [0]---container;
  [1]---volume; */
  string         division_time;    ///< "division time" 
  void get_positions_of_molecules(list<molecule>& vec_mol, size_t j,size_t mol_numb);
  parser (ifstream& init_input,ofstream* init_output);  /* constructor */
  ~parser(); /* destructor */ 
};  /* -----  end of class  parser  ----- */
/// constructor for parser
/** read input_file_name and give information about reagents and reactions. 
 *  Errors are written to output_file_name*/
parser::parser(ifstream& init_input,ofstream* init_output)
               : input(init_input),output(init_output),
                 colon(":"),plus("+"),eq("="),arrow(">"),comment(";"),comma(","),
                 num_mol("n"),diff_const("d"),react_const("k"),group("g"),
                 reagent("c"),reaction("r"),null_reagent("0"),
                 radius("rad"),sys("s"),position("pos"),volume("V"),division("division"),diff("d"),
                 result("o"),numb_exp("number_of_experiments"),time("time"),        
                 from_time("from_time"),numb_step_del("N"), 
                 line_number(1),exit_program(false)
{
  bool no_reagent=true;
  bool no_reaction=true;
  bool no_system=true;
  bool no_diffusive_part=true;
  bool diffusive=false;
  bool no_results=true;
  division_time="";
  while(getline(input,crude_line))
    {
      //prepare line
      erase_white_spaces();  
      erase_comments(); 
      check_if_empty(main_line);
      parse_line(crude_line,main_line,colon);
      //reagents and reactions
      if(!main_line.empty())
      {
        if(check_if(reagent))
          {
            no_reagent=false;
            bool if_mol_numb_0=false;
            check_if_empty(helper);
            if(get_mol_name(temp))   {helper.push_back(temp);}
            else                  {wr_comm("molecule name has not been found");exit_program=true;}
            if(get(num_mol,temp))   
            {
              helper.push_back(temp);
              if(temp=="0") if_mol_numb_0=true;             
            }
            else                  {wr_comm("number of molecules has not been found");exit_program=true;}
            if(get(diff_const,temp)) 
            {
              diffusive=true;
              helper.push_back(temp);
              if(get(radius,temp))  {helper.push_back(temp);}
              else                  {wr_comm("diffusion constant given but radius not!");exit_program=true;}
              if(get(position,temp))  helper.push_back(temp);
              else if(!if_mol_numb_0) {wr_comm("diffusion constant given but position not!");exit_program=true;}
            }
            reagents.push_back(helper);
          }
        if(check_if(reaction))
          {
            no_reaction=false;
            if(get_reaction()) 
              {
                check_if_empty(h);
                h.push_back(substr);
                h.push_back(produc);
                for(size_t i(0),n=reactions.size();i<n;++i)
                  {
                    if(substr==reactions[i][0] && produc==reactions[i][1])
                    {wr_comm("redefinition of reaction");exit_program=true;}
                  }
              }
            else               {wr_comm("reaction has not been found");exit_program=true;}
            if(get_react_const()) {h.push_back(k);}
            else               {wr_comm("reaction constant has not been found");exit_program=true;}
            reactions.push_back(h);
          }
        if(check_if(group))
        {
          bool diff_reag=false;
          for(size_t i(0);i<reagents.size();++i)
          {
             if(reagents[i].size() > 2) diff_reag=true;
          }
          if(!diff_reag) {wr_comm("there is group but there is not diffusive reagent");exit_program=true;}
          if(get_group())
            {
              groups_of_particles.push_back(helper);    
            }
          else
          {
            wr_comm("no group");
            exit_program=true;
          }
        }
        if(check_if(sys))
        {
          no_system=false;
          system.push_back(main_line[1]); //container
          if(get(volume,temp)) {system.push_back(temp);}
          else                  {wr_comm("no volume has been found");exit_program=true;}
        }          
        if(check_if(division))
        {
          if(get(time,temp)) {division_time=temp;}
        }
        if(check_if(diff))
        {
            no_diffusive_part=false;
            if(get_diffusive_reactions()) 
            {
              if(helper.size()!=2) {wr_comm("improperly written ,,diffusive reaction");exit_program=true;}
              if(helper[0]!=helper[1])
              {
                bool check_if_app_as_reac=false;
                for(size_t j(0),n=reactions.size();j<n;++j)
                  {
                    if(helper==reactions[j][0])
                    {check_if_app_as_reac=true;}
                  }
                if(!check_if_app_as_reac) {wr_comm("no definition for reaction");exit_program=true;}
              } 
              else
              {
                bool check_if_app_as_diff_reag=false;
                for(size_t j(0),n=reagents.size();j<n;++j)
                  {
                    if(helper[0]==reagents[j][0] && reagents[j].size()>2)
                    {check_if_app_as_diff_reag=true;}
                  }
                if(!check_if_app_as_diff_reag) 
                {
                  wr_comm("reagent "+helper[0]+" is not a diffusive reagent");exit_program=true;
                }
              }  
              diffusive_reactions.push_back(helper);    
            }
            else {wr_comm("no diffusive reaction");exit_program=true;}
        }
        if(check_if(result))
        {
          bool what_follow=false; 
          if(main_line[1]=="concentrations") {if(get_concentrations()) {measurement=h;what_follow=true;};}
          if(main_line[1]=="stopwatch") {if(get_stopwatch()) {measurement=h;what_follow=true;};}
      //    if(!what_follow)  {wr_comm("there isn't reagent to follow");exit_program=true;}
        }
      }  
      ++line_number;
    }
  --line_number;//the last line 
  if(no_reagent) {wr_comm("no reagent(s) has been found!");exit_program=true;}
  if(no_reaction) {wr_comm("no reaction(s) has been found!");exit_program=true;}
  if(no_system)   {wr_comm("no system has been found!");exit_program=true;}
  if(no_diffusive_part && diffusive)   
  {
    wr_comm("diffusive particles has been found but diffusive reactions not");exit_program=true;
  }
  if(exit_program) exit(1);
}  
/// destructor for parser
/** close input and output streams */
parser::~parser()
  {
  }
/**
 * @param str  --- a comment added to output file
 **/
void parser::wr_comm(string str)
  {
    (*output)<<"line:"<<line_number<<" "<<str<<endl;
  }
/**
 * @param str ---  a string for analysis
 * @param tokens  --- tokens
 * @param delimiters --- a delimiter
 **/
void parser::parse_line(const string& str, vector<string>& tokens, const string& delimiters)
    {
    // Skip delimiters at beginning.
    string::size_type lastPos = str.find_first_not_of(delimiters, 0);
    // Find first "non-delimiter".
    string::size_type pos = str.find_first_of(delimiters, lastPos);
    while (string::npos != pos || string::npos != lastPos)
       {
       // Found a token, add it to the vector.
       tokens.push_back(str.substr(lastPos, pos - lastPos));
       assert(tokens.back().size()>0);
       // Skip delimiters.  Note the "not_of"
       lastPos = str.find_first_not_of(delimiters, pos);
       // Find next "non-delimiter"
       pos = str.find_first_of(delimiters, lastPos);
       }
    }
/**
 * @param line  --- a checked vector 
 **/
void parser::check_if_empty(vector<string>& line)
  {
    if(!line.empty()) line.clear();
  }
void parser::check_if_empty(vector<vector<string> >& line)
  {
    if(!line.empty()) line.clear();
  }
/**
 * @return true if line with reagent has been found
 **/
bool parser::check_if(const string& sth)
  {
    if(main_line[0] == sth) return true;
    else {return false;}
  }
void parser::erase_white_spaces()
  {
    string white_space =" \t\r";
    string::size_type pos;
    while(true)
      {
        pos=crude_line.find_first_of(white_space);
        if(pos == string::npos) break;
        crude_line.erase(pos,1);
      }
  }
void parser::erase_comments()
  {
    string::size_type pos;
    pos=crude_line.find_first_of(comment);
    if(pos != string::npos)
      {
        crude_line.erase(pos);
      }
  }
/**
 * @param wht --- string with name of argument 
 * @param sth --- changed string 
 * @return true if argument has been found
 **/
bool parser::get(const string& wht,string& sth)
  {
    string::size_type pos;
    bool check=false;
    for(size_t i(1);i<main_line.size();i++)  //from 1 because 0 is information
      {
        pos=main_line[i].find_first_of(eq);
        if(pos != string::npos)
          {
            check_if_empty(inner_line);
            parse_line(main_line[i],inner_line,eq);
            if(inner_line[0] == wht) {sth=inner_line[1]; check=true;}
          }
      }
    return check;  
  }
/**
 * @param mol_name --- string with name of molecule
 * @return true if name of molecule has been found
 **/
bool parser::get_mol_name(string& mol_name)
  {
        string::size_type pos;
        pos=main_line[1].find_first_of(eq);
        if(pos == string::npos) 
          {
            mol_name=main_line[1]; 
            for(size_t i(0),n=reagents.size();i<n;++i)
              {
                if(reagents[i][0]==mol_name) {wr_comm(("redefinition of reagent: "+mol_name));exit_program=true;}
              }
            return true;
          }
        else                    {return false;}   
  }
bool parser::get_react_const()
  {
    check_if_empty(k);
//    string::size_type pos;
    bool check=false;
//    for(size_t i(1);i<main_line.size();i++)  //from 1 because 0 is information
//      {
//        pos=main_line[i].find_first_of(eq);
//        if(pos != string::npos)
//          {
//            check_if_empty(inner_line);
//            parse_line(main_line[i],inner_line,eq);
//            if(inner_line[0] == react_const) {k.push_back(inner_line[1]); check=true;}
//          }
//      } 
    check=get(react_const,temp);k.push_back(temp);
    if(get(numb_step_del,temp)) k.push_back(temp); //optional
    return check;  
  }
bool parser::get_reaction()
  {
    bool check=false;
    check_if_empty(substr);
    check_if_empty(produc);
    string::size_type pos,pos1;
    for(size_t i(1);i<main_line.size();i++)  //from 1 because 0 is information
      {
        pos=main_line[i].find_first_of(arrow);
        if(pos != string::npos)
          {
            check=true;
            check_if_empty(inner_line);
            parse_line(main_line[i],inner_line,arrow); 
            if(inner_line.size() < 2) {wr_comm("incorrectly written reaction");exit_program=true;}
            //substrates
            pos1=inner_line[0].find_first_of(plus);
            if(pos1 == string::npos)
            {
              bool appeared=false;
              for(size_t j(0),n=reagents.size();j<n;++j)
              {
                if(inner_line[0]==reagents[j][0] || inner_line[0]==null_reagent) 
                {appeared=true;}
              }
              if(!appeared) {wr_comm(("there is no definition for reagent: "+inner_line[0]));exit_program=true;}
              if(inner_line[0]!=null_reagent) substr.push_back(inner_line[0]);
            }
            else
            {
              check_if_empty(helper);
              parse_line(inner_line[0],helper,plus);
              for(size_t j(0);j<helper.size();j++)
                {
                  bool appeared=false;
                  for(size_t k(0),n=reagents.size();k<n;++k)
                    {
                      if(helper[j]==reagents[k][0]) 
                        {appeared=true;}
                    }
                  if(!appeared) {wr_comm(("there is no definition for substrate: "+helper[j]));exit_program=true;}
                  substr.push_back(helper[j]);
                }          
            }
            //products
            pos1=inner_line[1].find_first_of(plus);
            if(pos1 == string::npos)
            {
              bool appeared=false;
              for(size_t j(0),n=reagents.size();j<n;++j)
              {
                if(inner_line[1]==reagents[j][0] || inner_line[1]==null_reagent) 
                {appeared=true;}
              }
              if(!appeared) {wr_comm(("there is no definition for product: "+inner_line[1]));exit_program=true;}
              if(inner_line[1]!=null_reagent) produc.push_back(inner_line[1]);
            }
            else
            {
              check_if_empty(helper);
              parse_line(inner_line[1],helper,plus);
              for(size_t j(0);j<helper.size();j++)
                {
                  bool appeared=false;
                  for(size_t k(0),n=reagents.size();k<n;++k)
                    {
                      if(helper[j]==reagents[k][0]) 
                        {appeared=true;}
                    }
                  if(!appeared) {wr_comm(("there is no definition for product: "+helper[j]));exit_program=true;}
                  produc.push_back(helper[j]);
                }          
            }
          }
      }
    return check;
  }
bool parser::get_diffusive_reactions()
  {
    bool check=false;
    check_if_empty(helper);
    string::size_type pos;
    //substrates
    pos=main_line[1].find_first_of(plus);
    if(pos != string::npos)
      {
        check=true;
        check_if_empty(helper);
        parse_line(main_line[1],helper,plus);
        for(size_t j(0);j<helper.size();j++)
        {
          bool appeared=false;
          for(size_t k(0),n=reagents.size();k<n;++k)
          {
            if(helper[j]==reagents[k][0]) 
            {appeared=true;}
          }
          if(!appeared) 
          {
            wr_comm(("there is no definition for diffusive reagent: "+helper[j]));
            exit_program=true;
          }
        }          
      }
  }
bool parser::get_group()
  {
    bool check=false;
    check_if_empty(helper);
    string::size_type pos;
    //substrates
    if(main_line.size()>1)
    {
      check=true;
      pos=main_line[1].find_first_of(comma);
      if(pos == string::npos)
      { 
        bool appeared=false;
        for(size_t k(0),n=reagents.size();k<n;++k)
        {
          if(main_line[1]==reagents[k][0]) {appeared=true;helper.push_back(main_line[1]);}
        }
        if(!appeared) 
        {
          wr_comm("there is no definition for reagent in the group: "+main_line[1]);
          exit_program=true;
        }
      }
      else
      {
        check=true;
        parse_line(main_line[1],helper,comma);
        for(size_t j(0);j<helper.size();j++)
        {
          bool appeared=false;
          for(size_t k(0),n=reagents.size();k<n;++k)
          {
            if(helper[j]==reagents[k][0]) 
            {appeared=true;}
          }
          if(!appeared) 
          {
            wr_comm(("there is no definition for reagent in the group: "+helper[j]));
            exit_program=true;
          }
        }          
      }
    }
    return check;
  }
bool parser::get_concentrations()  
{ 
  bool appeared=false;
  check_if_empty(h);check_if_empty(helper);  
  helper.push_back(main_line[1]);
  if(get(numb_exp,temp)) 
  {
    check_if_empty(inner_line);
    parse_line(temp,inner_line,comma);
    if(inner_line.size()==0) 
    {  
      helper.push_back(temp);
    }
    for(size_t i(0);i<inner_line.size();++i)
    {
      helper.push_back(inner_line[i]);     
    }
  }
  else                  {wr_comm("no number of experiments has been found");exit_program=true;}
  h.push_back(helper);
  string opening_brace="{",closing_brace="}";
  while(getline(input,crude_line))
  {
    erase_white_spaces();  
    erase_comments(); 
    ++line_number;
    if(crude_line==opening_brace) continue;
    if(crude_line==closing_brace) break;
    check_if_empty(main_line);
    parse_line(crude_line,main_line,colon);
    bool reag_exist=false;
    for(size_t i(0);i<reagents.size();++i)
    {
      if(main_line[0]==reagents[i][0]) reag_exist=true;
    }
    check_if_empty(helper);
    if(reag_exist) {helper.push_back(main_line[0]);}
    else                  {wr_comm("reagent isn't declared: "+main_line[0]);exit_program=true;}
    if(!get(time,temp)) {wr_comm("what is the time of experiment?");exit_program=true;}
    check_if_empty(inner_line);
    parse_line(temp,inner_line,comma);
    for(size_t i(0);i<inner_line.size();++i)
    {
      helper.push_back(inner_line[i]);     
    }
    h.push_back(helper);
    appeared=true;
  }
  return appeared;
}
bool parser::get_stopwatch()
{
  bool appeared=false;
  check_if_empty(h);check_if_empty(helper);
  helper.push_back(main_line[1]);
  if(get(numb_exp,temp)) helper.push_back(temp);
  else                  {wr_comm("no number of experiments has been found");exit_program=true;}
  h.push_back(helper);
  string opening_brace="{",closing_brace="}";
  while(getline(input,crude_line))
  {
    erase_white_spaces();  
    erase_comments(); 
    ++line_number;
    if(crude_line==closing_brace) break;
    if(crude_line==opening_brace) continue;
    check_if_empty(main_line);
    parse_line(crude_line,main_line,colon);
    bool reag_exist=false;
    for(size_t i(0);i<reagents.size();++i)
    {
      if(main_line[0]==reagents[i][0]) reag_exist=true;
    }
    check_if_empty(helper);
    if(reag_exist) {helper.push_back(main_line[0]);}
    else                  {wr_comm("reagent has not been declared: "+main_line[0]);exit_program=true;}
    if(!get(num_mol,temp)) {wr_comm("the change of concentration is not given");exit_program=true;}
    helper.push_back(temp);
    h.push_back(helper);
    appeared=true;
  }
  return appeared;
}
void parser::get_positions_of_molecules(list<molecule>& list_mol, size_t j, size_t mol_numb)
{
  ifstream input_stream((reagents[j][4]).c_str());
  if(!input_stream) {(*output)<<"Error: file not found - "<<reagents[j][4]<<endl; exit(1);}
  bool if_e=false;
  size_t  line=0;
  while(getline(input_stream,crude_line))
  {
    check_if_empty(main_line);
    parse_line(crude_line,main_line);
    if(main_line.empty()) continue;  //skip empty lines
    if(main_line.size()!=3)	{(*output)<<"line: "<<line<<"Wrong position "<<" in "<<reagents[j][4]<<endl; if_e=true;}
    list_mol.push_back(molecule(
                               atof(main_line[0].c_str()),
                               atof(main_line[1].c_str()),
                               atof(main_line[2].c_str())
                              ));
    ++line;
  }
  if(line!=mol_numb) {(*output)<<reagents[j][4]<<" --- number of molecules is different from number of positions"<<endl; if_e=true;}
  if(if_e) exit(1);
}
