#include  "required.h"
#include  "reagent.cpp"
#include  "reaction.cpp"
#include  "connection.cpp"
#include  "vec.cpp"
#include  "molecule.cpp"
#include  "molecules.cpp"
#include  "diff_reaction.cpp"
#include  "container.cpp"
#include  "parser.cpp"
#include  "system.cpp"
struct n_e
{
 int     st_num_exp;
 int     end_numb_exp;
 int     curr_numb_exp;
 double  end_time_simulation;


 bool    concentrations;
 bool    stopwatch;
 n_e();
};
  n_e::n_e()                                                                               
      :st_num_exp(0),
       end_numb_exp(0),
       end_time_simulation(0),
       curr_numb_exp(st_num_exp),
       concentrations(false),
       stopwatch(false)
  {}
/**
 * =====================================================================================\n
 *        Class:  root\n
 *  Description:  a root of simulation\n
 * =====================================================================================
 */
class root : public connection
{
  public:
  vector<molecules*>        vect_molecules;                 ///<a vector of molecules
  vector<size_t>            vect_write_molecules;
  vector<size_t>            vect_init_molecules;
  vector<double>            vect_init_rrc;                

  vector<list<molecule> >   vect_group_molecules;           ///<contains a vector of molecules
  vector<diff_reaction>     vect_diff_reactions;            ///<reactions with diffusion                   
  container*                reactor;                        ///<reactor
  gsl_rng*                  rand_gen;                       ///<rng

  n_e                       info_exp;                       ///<info about numbers of experiments 
  string                    sim_dir;
  double                    div_time;
  bool                      division;
  file_management&          f_m;
  
  root (gsl_rng*, string&, string& , file_management&);  /* constructor */
  ~root();
  void reinitialize();
  void close_stream();
  private: 
  int get_ind_mol(const string&) const;            
  void initialize(ifstream& input,ofstream* output);
  string itos(int i, unsigned field_length=5);
  void initialize_ifstream();
};  


/**
 * @brief constructor
 * 
 * Initialize vector of reactions and vector of reagents from the class connection using class parser.
 * The default input file name is input.dat. Output file with errors from parsing is called input.log
 * 
 * @param input_file_name  --- file name with initial data for simulation
 *
 **/
root::root(gsl_rng* init_rand_gen,string& input, string& i_sim_dir, file_management& i_f_m)
     :rand_gen(init_rand_gen),sim_dir(i_sim_dir),f_m(i_f_m)
{
  ifstream input_stream(input.c_str());
  if(!input_stream) {*(f_m.output)<<"Error: file not found - "<<input<<endl; exit(1);}

  diff_molecules::put_rand(rand_gen);           //initialize rng in class molecules(static)
  initialize(input_stream,f_m.output);
  initialize_ifstream();
  input_stream.close();
}//root::root
root::~root()
{
  delete reactor;
  for(size_t i(0);i<vect_molecules.size();++i)
  {
    delete vect_molecules[i];
  }
}
string root::itos(int i, unsigned field_length)
//create string from the number(preceded by 0)
{
  ostringstream oss;
  oss << setfill('0') << setw(field_length) << i;
  return oss.str();
}//currently this methods work only with nrm 
/**
 * @brief get index of molecule
 *
 * @param str  --- name of molecule:)
 *
 * @return index
 **/
int root::get_ind_mol(const string& str) const
  {
    for(size_t i(0);i<vect_reagents.size();++i)
      {
        if(str == vect_reagents[i].name) return i;
      }
  }
void root::initialize(ifstream& input,ofstream* output)
  { 
    parser read_input_file(input,output);
    //initialize n_e
    if(read_input_file.measurement[0][0]=="concentrations")
    {
      if(read_input_file.measurement[0].size()==2) 
      {
        info_exp.end_numb_exp=atoi(read_input_file.measurement[0][1].c_str());
      }
      if(read_input_file.measurement[0].size()==3)
      {
        info_exp.st_num_exp=atoi(read_input_file.measurement[0][1].c_str());
        info_exp.curr_numb_exp=info_exp.st_num_exp;
        info_exp.end_numb_exp=atoi(read_input_file.measurement[0][2].c_str());
      }
      info_exp.concentrations=true;
    }
    else if(read_input_file.measurement[0][0]=="stopwatch")
    {
      info_exp.st_num_exp=0;
      info_exp.curr_numb_exp=0;
      info_exp.end_numb_exp=atoi(read_input_file.measurement[0][1].c_str());
      info_exp.stopwatch=true;
    }
    //initialize container
    if(read_input_file.system[0]=="sphere")
    {
      reactor=new sphere(atof(read_input_file.system[1].c_str()),rand_gen);
    }
    //initialize division time
    if(read_input_file.division_time!="")
    {
      division=true;
      div_time=atof(read_input_file.division_time.c_str());
    }
    else
    {
      division=false;
    }
    //reagents and molecules
    for(size_t i(0);i<read_input_file.reagents.size();i++)
      {
        add_reagent(read_input_file.reagents[i]);
        //ascribe a group
        for(size_t j(0);j<read_input_file.groups_of_particles.size();++j)
        {
         for(size_t k(0);k<read_input_file.groups_of_particles[j].size();++k)
         {
           if(read_input_file.reagents[i][0] == read_input_file.groups_of_particles[j][k])
           {
             vect_reagents[i].group=j;
           }
         }
        }
        //initialize molecules 	
        vect_init_molecules.push_back(atoi(read_input_file.reagents[i][1].c_str()));
        bool appeared=false; 
        for(size_t j(1);j<read_input_file.measurement.size();++j)
        {
          if(read_input_file.reagents[i][0] == read_input_file.measurement[j][0])
          {
            vect_write_molecules.push_back(i);
            if(info_exp.concentrations==true)
            {
              if(read_input_file.measurement[j].size()==4)
              {
                vect_molecules.push_back(new mfm_molecules_write_conc(atoi(read_input_file.reagents[i][1].c_str()),
                                                                      atof(read_input_file.measurement[j][1].c_str()),
                                                                      atof(read_input_file.measurement[j][2].c_str()),
                                                                      atof(read_input_file.measurement[j][3].c_str())
                                                                     ));
              }
              else if(read_input_file.measurement[j].size()==3)
              {
                vect_molecules.push_back(new mfm_molecules_write_conc(atoi(read_input_file.reagents[i][1].c_str()),
                                                                      atof(read_input_file.measurement[j][1].c_str()),
                                                                      atof(read_input_file.measurement[j][2].c_str())
                                                                     ));
              }
            }
            else if(info_exp.stopwatch==true)
            {
              vect_molecules.push_back(new mfm_molecules_write_stopwatch(atoi(read_input_file.reagents[i][1].c_str()),
                                                                         atoi(read_input_file.measurement[j][1].c_str())
                                                                        ));
            }
            appeared=true;
          } 
        } 
      if(!appeared)
        {
          if(read_input_file.reagents[i].size()>2)
          {
            vect_molecules.push_back(new diff_molecules(atoi(read_input_file.reagents[i][1].c_str())));
          }
          else
          {
            vect_molecules.push_back(new mfm_molecules(atoi(read_input_file.reagents[i][1].c_str())));
          }
        }
      }
    //reactions
    for(size_t i(0);i<read_input_file.reactions.size();i++)
      {
        add_reaction(atof(read_input_file.reactions[i][2][0].c_str()));
        vect_init_rrc.push_back(atof(read_input_file.reactions[i][2][0].c_str()));
        //delayed reactions
        if(read_input_file.reactions[i][2].size()==2) 
        {
          vect_reactions[i].delayed=true;
          vect_reactions[i].number_of_steps=atof(read_input_file.reactions[i][2][1].c_str());
        }
        for(size_t j(0);j<read_input_file.reactions[i][0].size();j++) //substrates
          {
            vect_reactions[i].add_subs_ind(get_ind_mol(read_input_file.reactions[i][0][j]));
          } 
        for(size_t j(0);j<read_input_file.reactions[i][1].size();j++) //products
          {
            vect_reactions[i].add_prod_ind(get_ind_mol(read_input_file.reactions[i][1][j]));
          } 
        //time for sorting:)
        //sort substrates---do it for 0dimensional 
        sort(vect_reactions[i].subs_ind.begin(),vect_reactions[i].subs_ind.end());    
        //check if substrates are simple;)
        for(size_t j(0);j<vect_reactions[i].subs_ind.size();++j)
          {
            if(count(vect_reactions[i].subs_ind.begin(),
                     vect_reactions[i].subs_ind.end(), 
                     vect_reactions[i].subs_ind[j]
                     ) > 1
              ) vect_reactions[i].no_simple=true;
          }
      }
    //let's return to reagents (initialization of as_substrate, as product)
    for(size_t i(0);i<vect_reagents.size();i++)
      {
        for(size_t j(0);j<vect_reactions.size();j++)
          {
            for(size_t k(0);k<vect_reactions[j].subs_ind.size();k++) 
              {
                if(i == vect_reactions[j].subs_ind[k]) vect_reagents[i].add_as_substrate(j);
              }  
            for(size_t k(0);k<vect_reactions[j].prod_ind.size();k++) 
              {
                if(i == vect_reactions[j].prod_ind[k]) vect_reagents[i].add_as_product(j);
              } 
          }
      }
    //initialize n_substrates n_products
    for(size_t i(0);i<vect_reagents.size();++i)
      {
        vect_reagents[i].n_substrates=vect_reagents[i].as_substrate.size();
        vect_reagents[i].n_products  =vect_reagents[i].as_product.size();
      }
    //initialize n_subs n_prods
    for(size_t i(0);i<vect_reactions.size();++i)
      {
        vect_reactions[i].n_subs=vect_reactions[i].subs_ind.size();
        vect_reactions[i].n_prods=vect_reactions[i].prod_ind.size();
      }
    //analyze diffusion reactions
    for(size_t i(0);i<read_input_file.diffusive_reactions.size();++i)   
    {
      size_t g1,g2;
      double D_a,D_b;
      double r_a,r_b;
      double i_k;
      //the species of the same kind
      if(read_input_file.diffusive_reactions[i][0]==read_input_file.diffusive_reactions[i][1])
      {
        bool appeared=false; //is it in the set of reactions?
        for(size_t j(0);j<read_input_file.reactions.size();j++) //substrates
        {
          if(read_input_file.diffusive_reactions[i]==read_input_file.reactions[j][0]) 
          {
            appeared=true;
            g1=vect_reagents[vect_reactions[j].subs_ind[0]].group;
            D_a=vect_reagents[vect_reactions[j].subs_ind[0]].diff_const;  
            r_a=vect_reagents[vect_reactions[j].subs_ind[0]].radius;   
            i_k=vect_reactions[j].react_rate_const;
            
            vect_diff_reactions.push_back(diff_reaction(g1,g1,D_a,D_a,r_a,r_a,i_k));
          }
        }
        if(!appeared)
        {
          for(size_t j(0);j<vect_reagents.size();j++)
          {
            if(read_input_file.diffusive_reactions[i][0]==vect_reagents[j].name)
            {
              g1=vect_reagents[j].group;
              D_a=vect_reagents[j].diff_const;
              r_a=vect_reagents[j].radius;
        
              vect_diff_reactions.push_back(diff_reaction(g1,D_a,r_a));
            }
          }
        }
      } 
      //two different species      
      else
      {
        bool appeared=false; //is it among reactions?
        for(size_t j(0);j<read_input_file.reactions.size();j++) //substrates
        {
          if(read_input_file.diffusive_reactions[i]==read_input_file.reactions[j][0]) 
          {
            appeared=true;
            g1=vect_reagents[vect_reactions[j].subs_ind[0]].group;
            g2=vect_reagents[vect_reactions[j].subs_ind[1]].group;
            D_a=vect_reagents[vect_reactions[j].subs_ind[0]].diff_const;  
            D_b=vect_reagents[vect_reactions[j].subs_ind[1]].diff_const;
            r_a=vect_reagents[vect_reactions[j].subs_ind[0]].radius;   
            r_b=vect_reagents[vect_reactions[j].subs_ind[1]].radius;
            i_k=vect_reactions[j].react_rate_const;
            
            vect_diff_reactions.push_back(diff_reaction(g1,g2,D_a,D_b,r_a,r_b,i_k));
          }
        }
      }
    }
    //the groups and positions of the molecules 
    string random="random";
    for(size_t i(0);i<read_input_file.groups_of_particles.size();++i)
    {
      list<molecule> helper;
      for(size_t j(0);j<vect_reagents.size();++j)
      {
        if(vect_reagents[j].group==i)
        {
          //this should be done after ascribing of reagents and molecules 
          if(vect_molecules[j]->mol_numb != 0 && read_input_file.reagents[j][4]==random)
          {
            for(size_t k(0);k<vect_molecules[j]->mol_numb;++k)
            {
              helper.push_back(molecule(reactor->random_initialize_in()));
            }
          }
          else
          {
            read_input_file.get_positions_of_molecules(helper,
                                                       j,
                                                       vect_molecules[j]->mol_numb
                                                      );
          }
        }
      } 
      vect_group_molecules.push_back(helper);
    }
    //initialize iterators tothe molecules
    for(size_t i(0);i<vect_group_molecules.size();++i)
    {
      list<molecule>::iterator iter_mol;
      iter_mol=vect_group_molecules[i].begin();
      for(size_t j(0);j<vect_reagents.size();++j)
      {
        if(vect_reagents[j].group==i)
        {
          diff_molecules* m=dynamic_cast<diff_molecules*>(vect_molecules[j]);
          for(size_t k(0);k<vect_molecules[j]->mol_numb;++k)
          {
            m->put_a_mol_iter(iter_mol);
            ++iter_mol;
          }
        }
      } 
    }
    for(size_t i(0);i<vect_reactions.size();++i)
    {
      size_t n=0;
      size_t m=vect_reactions[i].subs_ind.size();
      size_t k=vect_reactions[i].prod_ind.size();
      for(size_t j(0);j<m;++j)
      {
        if(vect_reagents[vect_reactions[i].subs_ind[j]].if_gill) ++n;
      }
      if(n)
      {
        if(n==m) vect_reactions[i].if_pure_g=true;
        else     vect_reactions[i].if_g=true;        
      }
      else 
      {
        bool prod_pure_g=true;
        for(size_t j(0);j<k;++j)
        {
          if(!vect_reagents[vect_reactions[i].prod_ind[j]].if_gill) prod_pure_g=false;
        }
        if(prod_pure_g && k!=0) vect_reactions[i].if_pure_g=true;
      }
      if(m==1 && !vect_reagents[vect_reactions[i].subs_ind[0]].if_gill)
      {
         vect_reactions[i].if_g=true;
      }
      if(m==0 || vect_reactions[i].if_pure_g)
      {
        for(size_t j(0);j<k;++j)
        {
          if(!vect_reagents[vect_reactions[i].prod_ind[j]].if_gill)  
          {
            vect_reactions[i].if_g=true;
            vect_reactions[i].if_pure_g=false;
          }
        }
      }
    } 
    //initialize n_e.end_time_simulation
    if(info_exp.concentrations)
    {
      mfm_molecules_write_conc* m_w_c; 
      for(size_t i(0);i<vect_write_molecules.size();++i)
      {
        m_w_c=dynamic_cast<mfm_molecules_write_conc*>(vect_molecules[vect_write_molecules[i]]);
        if(info_exp.end_time_simulation<(m_w_c->e_time)) info_exp.end_time_simulation=(m_w_c->e_time);
      }
    }
  } 
void root::initialize_ifstream()
{  
 for(size_t i(0),j(0);i<vect_write_molecules.size();++i)
 {
   j=vect_write_molecules[i];
   f_m.make_dir(sim_dir+vect_reagents[j].name);
   if(info_exp.concentrations)
   {
    (dynamic_cast<mfm_molecules_write_conc*>(vect_molecules[j]))->of
      =f_m.open_new_file(sim_dir+vect_reagents[j].name+"/"+itos(info_exp.curr_numb_exp));    
   }
   else if(info_exp.stopwatch)
   {
     (dynamic_cast<mfm_molecules_write_stopwatch*>(vect_molecules[j]))->of
       =f_m.open_new_file(sim_dir+vect_reagents[j].name+"/"+"times.dat");
   }
 }
}
void root::close_stream()
{
 for(size_t i(0),j(0);i<vect_write_molecules.size();++i)
 {
   if(info_exp.concentrations)
   {
    j=vect_write_molecules[i];
    mfm_molecules_write_conc* mfm=dynamic_cast<mfm_molecules_write_conc*>(vect_molecules[j]);
    f_m.close_file(mfm->of);
   }
   else if(info_exp.stopwatch)
   {
    j=vect_write_molecules[i];
    mfm_molecules_write_stopwatch* mfm=dynamic_cast<mfm_molecules_write_stopwatch*>(vect_molecules[j]);
    f_m.close_file(mfm->of);
   }
 }
}
void root::reinitialize()
{
 if(info_exp.concentrations) 
 {
   for(size_t i(0),j(0);i<vect_write_molecules.size();++i)
   {
     j=vect_write_molecules[i];
     mfm_molecules_write_conc* mfm=dynamic_cast<mfm_molecules_write_conc*>(vect_molecules[j]);
     f_m.close_file(mfm->of);
     mfm->of=f_m.open_new_file(sim_dir+vect_reagents[j].name+"/"+itos(info_exp.curr_numb_exp));    
     mfm->p_time=mfm->s_time;  
   }
 }
 for(size_t i(0);i<vect_molecules.size();++i)
 {
   vect_molecules[i]->mol_numb=vect_init_molecules[i];
 }
 //for(size_t i(0);i<vect_reactions.size();++i)
 //{
 //  vect_reactions[i].react_rate_const=vect_init_rrc[i];
// }

}
