#include<signal.h>
class simulation
{
  public:
  double time;                     
  double prev_time;
  double d_t;  
  reaction*      r; 
  simulation (root& ,gsl_rng*);                
  ~simulation ();                 
  public:
  root&                   rt;                  
  gsl_rng*              rand_gen;               //generator of random numbers 
  NEXT_REACTION_METHOD*   nrm;                  //0dimensional method 
  bool check_reactions();
  void react();
  void change_concentrations();
  bool if_nrm();
  void ascribe();
  void simulate();
};
simulation::simulation (root& init_root,gsl_rng* init_rand_gen)
          :rt(init_root),
          rand_gen(init_rand_gen),
          time(0.0),r(NULL),d_t(100.0)
{
  ascribe();
  if(if_nrm()) nrm=new NEXT_REACTION_METHOD(rt,rand_gen);  
  if(rt.info_exp.concentrations)
  {
    for( ; ; )
    {
       time=0.0;
       d_t=rt.div_time;
       for(size_t i(0),j(0);i<rt.vect_write_molecules.size();++i)
           {
             j=rt.vect_write_molecules[i];
             (dynamic_cast<mfm_molecules_write_conc*>(rt.vect_molecules[j]))->write_initial_state();
           }
       while(1)
       {
         if(check_reactions())   
         { 
            continue;  
         }
         change_concentrations();   //change number of molecules and write state
         react();                   //make reaction
         if(time>rt.info_exp.end_time_simulation) 
         {
           for(size_t i(0),j(0);i<rt.vect_write_molecules.size();++i)
           {
             j=rt.vect_write_molecules[i];
             (dynamic_cast<mfm_molecules_write_conc*>(rt.vect_molecules[j]))->complete_state();
           }
           break;
         }
       }
       ++rt.info_exp.curr_numb_exp;
       if(rt.info_exp.curr_numb_exp<rt.info_exp.end_numb_exp) 
       {
         rt.reinitialize();
         nrm->reinitialize();   
       }
       else
       {
         rt.close_stream();
         break;
       }
    }
  }
  else if(rt.info_exp.stopwatch)
  {
    bool FIN;
    for( ; ; )
    {
      time=0.0;
      while(1)
      {
        FIN=true;
        check_reactions();
        change_concentrations();
        for(size_t i(0),j(0);i<rt.vect_write_molecules.size();++i)
        {
          j=rt.vect_write_molecules[i];
          if(!(dynamic_cast<mfm_molecules_write_stopwatch*>(rt.vect_molecules[j]))->finish)
          {
            FIN=false;
          }
        }
        if( FIN )
        {
          break;           
        }
      react();      
      }
      ++rt.info_exp.curr_numb_exp;
      if(rt.info_exp.curr_numb_exp==rt.info_exp.end_numb_exp) 
      {
         rt.close_stream();
         break;
      }
      //reinitialization
      rt.reinitialize();
      nrm->reinitialize(); 
      for(size_t i(0),j(0);i<rt.vect_write_molecules.size();++i)
      {
        j=rt.vect_write_molecules[i];
        (dynamic_cast<mfm_molecules_write_stopwatch*>(rt.vect_molecules[j]))->finish=false;
      }
    }
  }
}  /* -----  end of method simulation::simulation  (constructor)  ----- */
simulation::~simulation ()
{

}  /* -----  end of method simulation::~simulation  (destructor)  ----- */
bool simulation::check_reactions()
{
  prev_time=time;
  nrm->check_reaction(time,r);
  if(rt.division)
  {
    if(prev_time<= d_t && time >d_t)
    {
      int ind_jump_reac=0;
      nrm->division(d_t);
      for(int i(0),j(0);i<rt.vect_write_molecules.size();++i)
      {
        j=rt.vect_write_molecules[i];
        if(!(dynamic_cast<mfm_molecules_write_conc*>(rt.vect_molecules[j]))->with_step)
        {
         time=d_t;
         (dynamic_cast<mfm_molecules_write_conc*>(rt.vect_molecules[j]))->write_state();
         nrm->check_reaction(time,r);
        }
      }
    d_t+=rt.div_time;;    
    return true;
    }
  //if(prev_time<= 15000.0 && time >15000.0)
  //{
  //  nrm->rrc_jump(0.0955,8,15000.0);
  //  return true;
  //}
    else
    {
    return false;
    }
  }
  return false;
}
void simulation::change_concentrations()
{
  for(size_t i(0);i<r->n_subs;++i)
  {
    rt.vect_molecules[r->subs_ind[i]]->minus_mol();
  }
  for(size_t i(0);i<r->n_prods;++i)
  {
    rt.vect_molecules[r->prod_ind[i]]->plus_mol();
  }
}
void simulation::react()
{
  nrm->react();    
}  
bool simulation::if_nrm()
{
  for(size_t i(0);i<rt.vect_reactions.size();++i)
  {
    if(rt.vect_reactions[i].if_g || rt.vect_reactions[i].if_pure_g) {return true;}      
  }
  return false;
}
void simulation::ascribe()
  {
    save_state::time=&time;   
  }
